#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <asm/types.h>
#ifdef USE_MTD_USER
# include <mtd/mtd-user.h>
#else
struct mtd_oob_buf {
	uint32_t start;
	uint32_t length;
	unsigned char *ptr;
};
struct mtd_info_user {
	uint8_t type;
	uint32_t flags;
	uint32_t size;
	uint32_t erasesize;
	uint32_t oobblock;
	uint32_t oobsize;
	uint32_t ecctype;
	uint32_t eccsize;
};
# define MEMGETINFO _IOR('M', 1, struct mtd_info_user)
# define MEMREADOOB _IOWR('M', 4, struct mtd_oob_buf)
# define MEMGETBADBLOCK _IOW('M', 11, loff_t)
#endif

struct info {
	int fd;
	uint32_t size;
	uint32_t erasesize;
	int eraseblocks;
	uint32_t oobblock;
	int pages;
	int pages_pr_block;
	uint32_t oobsize;
	int pageoobsize;
	int goodblocks;
	char name[32];
} info[3];

int debug = 0;

void die(const char *msg)
{
	perror(msg);
	exit(1);
}

int is_bad(struct info *f, loff_t offs)
{
	int flag = ioctl(f->fd, MEMGETBADBLOCK, &offs);
	if (flag < 0)
		die("Ioctl MEMGETBADBLOCK");
	return flag;
}

int count_good(struct info *f)
{
	int count = 0;
	int i;
	for (i = 0; i < f->eraseblocks; i++) {
		if (!is_bad(f, i * f->erasesize))
			++count;
		else if (debug)
			fprintf(stderr, "Bad block found at eraseblock %d (0x%x)\n", i, i * f->erasesize);
		
	}
	return count;
}

void open_mtd(const char *name, struct info *f)
{
	struct mtd_info_user minfo;
	strncpy(f->name, name, 31);
	f->fd = open(name, O_RDONLY);
	if (f->fd < 0) die("open mtd");
	if (ioctl(f->fd, MEMGETINFO, &minfo) < 0) die("Ioctl MEMGETINFO");
	f->size = minfo.size;
	f->erasesize = minfo.erasesize;
	f->eraseblocks = f->size / f->erasesize;
	f->oobblock = minfo.oobblock;
	f->pages = f->size / f->oobblock;
	f->pages_pr_block = f->erasesize / f->oobblock;
	f->oobsize = minfo.oobsize;
	f->pageoobsize = f->oobblock + f->oobsize;
	if (debug) {
		fprintf(stderr, "%s information:\n", f->name);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "Size", f->size, f->size);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "Erasesize", f->erasesize, f->erasesize);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "Eraseblocks", f->eraseblocks, f->eraseblocks);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "Pagesize", f->oobblock, f->oobblock);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "Pages", f->pages, f->pages);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "Pages_pr_block", f->pages_pr_block, f->pages_pr_block);
		fprintf(stderr, "\t%-16s : %d (0x%x)\n", "OObsize", f->oobsize, f->oobsize);
		fprintf(stderr, "\nScanning for bad blocks\n");
	}
	f->goodblocks = count_good(f);
	if (debug)
		fprintf(stderr, "%d bad blocks found\n", f->eraseblocks - f->goodblocks);
}

unsigned char readbuf[2048];
unsigned char oobdata[64];
char file_header[32] = "NFI1";

void read_oob(struct info *f, loff_t offs)
{
	struct mtd_oob_buf oob = {offs, f->oobsize, oobdata};
	if (ioctl(f->fd, MEMREADOOB, &oob) < 0)
		die("Ioctl MEMREADOOB");
}

void emit_4(unsigned long val)
{
	val = htonl(val);
	write(1, &val, 4);
}

void emit_file(struct info *f)
{
	int i, j;
	if (debug)
		fprintf(stderr, "Dumping blocks for %s\n", f->name);
	emit_4(f->goodblocks * f->pageoobsize * f->pages_pr_block);
	for (i = 0; i < f->eraseblocks; i++) {
		lseek(f->fd, i * f->erasesize, SEEK_SET);
		if (is_bad(f, i * f->erasesize)) {
			if (debug)
				fprintf(stderr, "Skipping bad block at 0x%x\n", i * f->erasesize);
			continue;
		}
		for (j = 0; j < f->pages_pr_block; j++) {
			if (read(f->fd, readbuf, f->oobblock) < 0)
				die("Reading mtd");
			if (write(1, readbuf, f->oobblock) < 0)
				die("Writing page");
			read_oob(f, i * f->erasesize + j * f->oobblock);
			if (write(1, oobdata, f->oobsize) < 0)
				die("Writing oob");
		}
	}
}

int main(int argc, char **argv)
{
	int i;
	if (argc > 1 && (strcmp(argv[1], "-v") == 0)) {
		debug = 1;
		--argc;
		++argv;
	}
	if (argc == 2) {
		strncat(file_header, argv[1], 28);
		if (strcmp(argv[1], "dm7025") == 0)
			write(1, file_header, 32);
	} else {
		fprintf(stderr, "Usage: backup-dreambox dm7020|dm7025\n");
		exit(1);
	}

	open_mtd("/dev/mtd/1ro", &info[0]);
	open_mtd("/dev/mtd/2ro", &info[1]);
	open_mtd("/dev/mtd/3ro", &info[2]);

	int grand_total = 4 + 3 * 4;
	for (i = 0; i < 3; i++) {
		struct info *f = &info[i];
		grand_total += 4 + f->goodblocks * f->pageoobsize * f->pages_pr_block;
	}

	emit_4(grand_total);
	emit_4(3 * 4);
	emit_4(0x40000);
	emit_4(0x400000);
	emit_4(0x2000000);

	emit_file(&info[0]);
	emit_file(&info[1]);
	emit_file(&info[2]);
}
